@@ -1,28 +1,39 @@
+from __future__ import annotations
+
 import codecs
 import collections
+import difflib
 import sys
 import typing as t
 
 from ._common import colorize
 
 
-def _gen_change_caret_line(
-    original, updated, charwidth: t.Optional[t.Callable[[str], int]]
-):
-    shift = 0
-    indices = []
-    for idx, c in enumerate(original):
-        if c != updated[idx + shift]:
-            indices.append(idx)
-            if charwidth is not None:
-                shift += charwidth(c) - 1
-    gen = ""
-    cur = 0
-    for idx in indices:
-        gen += " " * (idx - cur)
-        gen += "^"
-        cur = idx
-    return gen
+def create_comparison_lines(old: str, new: str) -> list[str]:
+    """Compare two lines to make diff output to show changes."""
+    differ = difflib.Differ()
+    return [_clean_q(line).rstrip("\n") for line in differ.compare([old], [new])]
+
+
+# simplify the format produced by Differ (for single-line comparison)
+def _clean_q(s: str) -> str:
+    if s.startswith("?"):
+        return " " + s[1:]
+    return s
+
+
+def colorize_comparison(comparison: list[str]) -> list[str]:
+    result = []
+    for line in comparison:
+        if line.startswith("-"):
+            result.append(colorize(line, color="bright_red"))
+        elif line.startswith("+"):
+            result.append(colorize(line, color="bright_green"))
+        elif line.startswith("?"):
+            result.append(colorize(line, color="bright_cyan"))
+        else:
+            result.append(line)
+    return result
 
 
 def _determine_encoding() -> str:
@@ -44,7 +55,7 @@ def _readlines(filename: str, encoding: str) -> t.List[str]:
 
 
 class _VPrinter:
-    def __init__(self, verbosity: int):
+    def __init__(self, verbosity: int) -> None:
         self.verbosity = verbosity
 
     def out(self, message: str, verbosity: int = 1, end: str = "\n") -> None:
@@ -54,14 +65,14 @@ class _VPrinter:
 
 
 class DiffRecorder:
-    def __init__(self, verbosity: int):
+    def __init__(self, verbosity: int) -> None:
         self._printer = _VPrinter(verbosity)
         # in py3.6+ the dict builtin maintains order, but being explicit is
         # slightly safer since we're being explicit about the fact that we want
         # to retain key order
-        self.by_fname: t.MutableMapping[
-            str, t.List[t.Tuple[str, str, int]]
-        ] = collections.OrderedDict()
+        self.by_fname: t.MutableMapping[str, t.List[t.Tuple[str, str, int]]] = (
+            collections.OrderedDict()
+        )
         self._file_encoding = _determine_encoding()
 
     def add(self, fname, original, updated, lineno):
@@ -84,11 +95,18 @@ class DiffRecorder:
 
         Returns True if changes were made, False if none were made"""
         self._printer.out(f"checking {filename}...", end="", verbosity=2)
-        content = _readlines(filename, self._file_encoding)
+        try:
+            content = _readlines(filename, self._file_encoding)
+        except FileNotFoundError:
+            self._printer.out(f"fail, FileNotFound: {filename}", verbosity=1)
+            raise
 
         newcontent = []
         for lineno, line in enumerate(content, 1):
             newline = line_fixer(line)
+            # re-add newline if it was stripped by the fixer
+            if line.endswith("\n") and not newline.endswith("\n"):
+                newline += "\n"
             newcontent.append(newline)
             if newline != line:
                 self.add(filename, line, newline, lineno)
@@ -116,29 +134,21 @@ class DiffRecorder:
                 filename_c = filename
             self._printer.out(f"  {filename_c}")
             if show_changes:
-                for (original, updated, lineno) in changeset:
-                    original = "-" + original.rstrip()
-                    updated = "+" + updated.rstrip()
-                    caret_line = " " + _gen_change_caret_line(
-                        original[1:], updated[1:], charwidth
-                    )
-
+                for original, updated, lineno in changeset:
+                    comparison = create_comparison_lines(original, updated)
                     if ansi_colors:
-                        original = colorize(original, color="bright_red")
-                        updated = colorize(updated, color="bright_green")
-                        caret_line = colorize(caret_line, color="bright_cyan")
+                        comparison = colorize_comparison(comparison)
                     self._printer.out(f"  line {lineno}:")
-                    self._printer.out(f"    {original}")
-                    self._printer.out(f"    {updated}")
-                    self._printer.out(f"    {caret_line}")
+                    for line in comparison:
+                        self._printer.out(f"    {line}")
 
 
 class CheckRecorder:
-    def __init__(self, verbosity: int):
+    def __init__(self, verbosity: int) -> None:
         self._printer = _VPrinter(verbosity)
-        self.by_fname: t.MutableMapping[
-            str, t.List[t.Tuple[str, str, int]]
-        ] = collections.OrderedDict()
+        self.by_fname: t.MutableMapping[str, t.List[t.Tuple[str, str, int]]] = (
+            collections.OrderedDict()
+        )
         self._file_encoding = _determine_encoding()
 
     def add(self, fname, lineno):
